/* -*-c-*-
 *
 * $Id: mpx.c,v 1.3 1999/11/13 01:57:31 mdw Exp $
 *
 * Low-level multiprecision arithmetic
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: mpx.c,v $
 * Revision 1.3  1999/11/13 01:57:31  mdw
 * Remove stray debugging code.
 *
 * Revision 1.2  1999/11/13 01:50:59  mdw
 * Multiprecision routines finished and tested.
 *
 * Revision 1.1  1999/09/03 08:41:12  mdw
 * Initial import.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mLib/bits.h>

#include "mptypes.h"
#include "mpx.h"

/*----- Loading and storing -----------------------------------------------*/

/* --- @mpx_storel@ --- *
 *
 * Arguments:	@const mpw *v, *vl@ = base and limit of source vector
 *		@void *pp@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Stores an MP in an octet array, least significant octet
 *		first.  High-end octets are silently discarded if there
 *		isn't enough space for them.
 */

void mpx_storel(const mpw *v, const mpw *vl, void *pp, size_t sz)
{
  mpw n, w = 0;
  octet *p = pp, *q = p + sz;
  unsigned bits = 0;

  while (p < q) {
    if (bits < 8) {
      if (v >= vl) {
	*p++ = U8(w);
	break;
      }
      n = *v++;
      *p++ = U8(w | n << bits);
      w = n >> (8 - bits);
      bits += MPW_BITS - 8;
    } else {
      *p++ = U8(w);
      w >>= 8;
      bits -= 8;
    }
  }
  memset(p, 0, q - p);
}

/* --- @mpx_loadl@ --- *
 *
 * Arguments:	@mpw *v, *vl@ = base and limit of destination vector
 *		@const void *pp@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Loads an MP in an octet array, least significant octet
 *		first.  High-end octets are ignored if there isn't enough
 *		space for them.
 */

void mpx_loadl(mpw *v, mpw *vl, const void *pp, size_t sz)
{
  unsigned n;
  mpw w = 0;
  const octet *p = pp, *q = p + sz;
  unsigned bits = 0;

  if (v >= vl)
    return;
  while (p < q) {
    n = U8(*p++);
    w |= n << bits;
    bits += 8;
    if (bits >= MPW_BITS) {
      *v++ = MPW(w);
      w = n >> (MPW_BITS - bits + 8);
      bits -= MPW_BITS;
      if (v >= vl)
	return;
    }
  }
  *v++ = w;
  MPX_ZERO(v, vl);
}

/* --- @mpx_storeb@ --- *
 *
 * Arguments:	@const mpw *v, *vl@ = base and limit of source vector
 *		@void *pp@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Stores an MP in an octet array, most significant octet
 *		first.  High-end octets are silently discarded if there
 *		isn't enough space for them.
 */

void mpx_storeb(const mpw *v, const mpw *vl, void *pp, size_t sz)
{
  mpw n, w = 0;
  octet *p = pp, *q = p + sz;
  unsigned bits = 0;

  while (q > p) {
    if (bits < 8) {
      if (v >= vl) {
	*--q = U8(w);
	break;
      }
      n = *v++;
      *--q = U8(w | n << bits);
      w = n >> (8 - bits);
      bits += MPW_BITS - 8;
    } else {
      *--q = U8(w);
      w >>= 8;
      bits -= 8;
    }
  }
  memset(p, 0, q - p);
}

/* --- @mpx_loadb@ --- *
 *
 * Arguments:	@mpw *v, *vl@ = base and limit of destination vector
 *		@const void *pp@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Loads an MP in an octet array, most significant octet
 *		first.  High-end octets are ignored if there isn't enough
 *		space for them.
 */

void mpx_loadb(mpw *v, mpw *vl, const void *pp, size_t sz)
{
  unsigned n;
  mpw w = 0;
  const octet *p = pp, *q = p + sz;
  unsigned bits = 0;

  if (v >= vl)
    return;
  while (q > p) {
    n = U8(*--q);
    w |= n << bits;
    bits += 8;
    if (bits >= MPW_BITS) {
      *v++ = MPW(w);
      w = n >> (MPW_BITS - bits + 8);
      bits -= MPW_BITS;
      if (v >= vl)
	return;
    }
  }
  *v++ = w;
  MPX_ZERO(v, vl);
}

/*----- Logical shifting --------------------------------------------------*/

/* --- @mpx_lsl@ --- *
 *
 * Arguments:	@mpw *dv, *dvl@ = destination vector base and limit
 *		@const mpw *av, *avl@ = source vector base and limit
 *		@size_t n@ = number of bit positions to shift by
 *
 * Returns:	---
 *
 * Use:		Performs a logical shift left operation on an integer.
 */

void mpx_lsl(mpw *dv, mpw *dvl, const mpw *av, const mpw *avl, size_t n)
{
  size_t nw;
  unsigned nb;

  /* --- Trivial special case --- */

  if (n == 0)
    MPX_COPY(dv, dvl, av, avl);

  /* --- Single bit shifting --- */

  else if (n == 1) {
    mpw w = 0;
    while (av < avl) {
      mpw t;
      if (dv >= dvl)
	goto done;
      t = *av++;
      *dv++ = MPW((t << 1) | w);
      w = t >> (MPW_BITS - 1);
    }
    if (dv >= dvl)
      goto done;
    *dv++ = MPW(w);
    MPX_ZERO(dv, dvl);
    goto done;
  }

  /* --- Break out word and bit shifts for more sophisticated work --- */
	
  nw = n / MPW_BITS;
  nb = n % MPW_BITS;

  /* --- Handle a shift by a multiple of the word size --- */

  if (nb == 0) {
    MPX_COPY(dv + nw, dvl, av, avl);
    memset(dv, 0, MPWS(nw));
  }

  /* --- And finally the difficult case --- *
   *
   * This is a little convoluted, because I have to start from the end and
   * work backwards to avoid overwriting the source, if they're both the same
   * block of memory.
   */

  else {
    mpw w;
    size_t nr = MPW_BITS - nb;
    size_t dvn = dvl - dv;
    size_t avn = avl - av;

    if (dvn <= nw) {
      MPX_ZERO(dv, dvl);
      goto done;
    }

    if (dvn > avn + nw) {
      size_t off = avn + nw + 1;
      MPX_ZERO(dv + off, dvl);
      dvl = dv + off;
      w = 0;
    } else {
      avl = av + dvn - nw;
      w = *--avl << nb;
    }

    while (avl > av) {
      mpw t = *--avl;
      *--dvl = (t >> nr) | w;
      w = t << nb;
    }

    *--dvl = w;
    MPX_ZERO(dv, dvl);
  }

done:;
}

/* --- @mpx_lsr@ --- *
 *
 * Arguments:	@mpw *dv, *dvl@ = destination vector base and limit
 *		@const mpw *av, *avl@ = source vector base and limit
 *		@size_t n@ = number of bit positions to shift by
 *
 * Returns:	---
 *
 * Use:		Performs a logical shift right operation on an integer.
 */

void mpx_lsr(mpw *dv, mpw *dvl, const mpw *av, const mpw *avl, size_t n)
{
  size_t nw;
  unsigned nb;

  /* --- Trivial special case --- */

  if (n == 0)
    MPX_COPY(dv, dvl, av, avl);

  /* --- Single bit shifting --- */

  else if (n == 1) {
    mpw w = *av++ >> 1;
    while (av < avl) {
      mpw t;
      if (dv >= dvl)
	goto done;
      t = *av++;
      *dv++ = MPW((t << (MPW_BITS - 1)) | w);
      w = t >> 1;
    }
    if (dv >= dvl)
      goto done;
    *dv++ = MPW(w);
    MPX_ZERO(dv, dvl);
    goto done;
  }

  /* --- Break out word and bit shifts for more sophisticated work --- */

  nw = n / MPW_BITS;
  nb = n % MPW_BITS;

  /* --- Handle a shift by a multiple of the word size --- */

  if (nb == 0)
    MPX_COPY(dv, dvl, av + nw, avl);

  /* --- And finally the difficult case --- */

  else {
    mpw w;
    size_t nr = MPW_BITS - nb;

    av += nw;
    w = *av++;
    while (av < avl) {
      mpw t;
      if (dv >= dvl)
	goto done;
      t = *av++;
      *dv++ = MPW((w >> nb) | (t << nr));
      w = t;
    }
    if (dv < dvl) {
      *dv++ = MPW(w >> nb);
      MPX_ZERO(dv, dvl);
    }
  }

done:;
}

/*----- Unsigned arithmetic -----------------------------------------------*/

/* --- @mpx_ucmp@ --- *
 *
 * Arguments:	@const mpw *av, *avl@ = first argument vector base and limit
 *		@const mpw *bv, *bvl@ = second argument vector base and limit
 *
 * Returns:	Less than, equal to, or greater than zero depending on
 *		whether @a@ is less than, equal to or greater than @b@,
 *		respectively.
 *
 * Use:		Performs an unsigned integer comparison.
 */

int mpx_ucmp(const mpw *av, const mpw *avl, const mpw *bv, const mpw *bvl)
{
  MPX_SHRINK(av, avl);
  MPX_SHRINK(bv, bvl);

  if (avl - av > bvl - bv)
    return (+1);
  else if (avl - av < bvl - bv)
    return (-1);
  else while (avl > av) {
    mpw a = *--avl, b = *--bvl;
    if (a > b)
      return (+1);
    else if (a < b)
      return (-1);
  }
  return (0);
}
  
/* --- @mpx_uadd@ --- *
 *
 * Arguments:	@mpw *dv, *dvl@ = destination vector base and limit
 *		@const mpw *av, *avl@ = first addend vector base and limit
 *		@const mpw *bv, *bvl@ = second addend vector base and limit
 *
 * Returns:	---
 *
 * Use:		Performs unsigned integer addition.  If the result overflows
 *		the destination vector, high-order bits are discarded.  This
 *		means that two's complement addition happens more or less for
 *		free, although that's more a side-effect than anything else.
 *		The result vector may be equal to either or both source
 *		vectors, but may not otherwise overlap them.
 */

void mpx_uadd(mpw *dv, mpw *dvl, const mpw *av, const mpw *avl,
	      const mpw *bv, const mpw *bvl)
{
  mpw c = 0;

  while (av < avl || bv < bvl) {
    mpw a, b;
    mpd x;
    if (dv >= dvl)
      return;
    a = (av < avl) ? *av++ : 0;
    b = (bv < bvl) ? *bv++ : 0;
    x = (mpd)a + (mpd)b + c;
    *dv++ = MPW(x);
    c = x >> MPW_BITS;
  }
  if (dv < dvl) {
    *dv++ = c;
    MPX_ZERO(dv, dvl);
  }
}

/* --- @mpx_usub@ --- *
 *
 * Arguments:	@mpw *dv, *dvl@ = destination vector base and limit
 *		@const mpw *av, *avl@ = first argument vector base and limit
 *		@const mpw *bv, *bvl@ = second argument vector base and limit
 *
 * Returns:	---
 *
 * Use:		Performs unsigned integer subtraction.  If the result
 *		overflows the destination vector, high-order bits are
 *		discarded.  This means that two's complement subtraction
 *		happens more or less for free, althuogh that's more a side-
 *		effect than anything else.  The result vector may be equal to
 *		either or both source vectors, but may not otherwise overlap
 *		them.
 */

void mpx_usub(mpw *dv, mpw *dvl, const mpw *av, const mpw *avl,
	      const mpw *bv, const mpw *bvl)
{
  mpw c = 0;

  while (av < avl || bv < bvl) {
    mpw a, b;
    mpd x;
    if (dv >= dvl)
      return;
    a = (av < avl) ? *av++ : 0;
    b = (bv < bvl) ? *bv++ : 0;
    x = (mpd)a - (mpd)b - c;
    *dv++ = MPW(x);
    if (x >> MPW_BITS)
      c = 1;
    else
      c = 0;
  }
  if (c)
    c = MPW_MAX;
  while (dv < dvl)
    *dv++ = c;
}

/* --- @mpx_umul@ --- *
 *
 * Arguments:	@mpw *dv, *dvl@ = destination vector base and limit
 *		@const mpw *av, *avl@ = multiplicand vector base and limit
 *		@const mpw *bv, *bvl@ = multiplier vector base and limit
 *
 * Returns:	---
 *
 * Use:		Performs unsigned integer multiplication.  If the result
 *		overflows the desination vector, high-order bits are
 *		discarded.  The result vector may not overlap the argument
 *		vectors in any way.
 */

void mpx_umul(mpw *dv, mpw *dvl, const mpw *av, const mpw *avl,
	      const mpw *bv, const mpw *bvl)
{
  /* --- This is probably worthwhile on a multiply --- */

  MPX_SHRINK(av, avl);
  MPX_SHRINK(bv, bvl);

  /* --- Deal with a multiply by zero --- */
  
  if (bv == bvl) {
    MPX_ZERO(dv, dvl);
    return;
  }

  /* --- Do the initial multiply and initialize the accumulator --- */

  MPX_UMULN(dv, dvl, av, avl, *bv++);

  /* --- Do the remaining multiply/accumulates --- */

  while (dv < dvl && bv < bvl) {
    mpw m = *bv++;
    mpw c = 0;
    const mpw *avv = av;
    mpw *dvv = ++dv;

    while (avv < avl) {
      mpd x;
      if (dvv >= dvl)
	goto next;
      x = (mpd)*dvv + (mpd)m * (mpd)*avv++ + c;
      *dvv++ = MPW(x);
      c = x >> MPW_BITS;
    }
    MPX_UADDN(dvv, dvl, c);
  next:;
  }
}

/* --- @mpx_usqr@ --- *
 *
 * Arguments:	@mpw *dv, *dvl@ = destination vector base and limit
 *		@const mpw *av, *av@ = source vector base and limit
 *
 * Returns:	---
 *
 * Use:		Performs unsigned integer squaring.  The result vector must
 *		not overlap the source vector in any way.
 */

void mpx_usqr(mpw *dv, mpw *dvl, const mpw *av, const mpw *avl)
{
  MPX_ZERO(dv, dvl);

  /* --- Main loop --- */

  while (av < avl) {
    const mpw *avv = av;
    mpw *dvv = dv;
    mpw a = *av;
    mpd c;

    /* --- Stop if I've run out of destination --- */

    if (dvv >= dvl)
      break;

    /* --- Work out the square at this point in the proceedings --- */

    {
      mpd x = (mpd)a * (mpd)a + *dvv;
      *dvv++ = MPW(x);
      c = MPW(x >> MPW_BITS);
    }

    /* --- Now fix up the rest of the vector upwards --- */

    avv++;
    while (dvv < dvl && avv < avl) {
      mpd x = (mpd)a * (mpd)*avv++;
      mpd y = ((x << 1) & MPW_MAX) + c + *dvv;
      c = (x >> (MPW_BITS - 1)) + (y >> MPW_BITS);
      *dvv++ = MPW(y);
    }
    while (dvv < dvl && c) {
      mpd x = c + *dvv;
      *dvv++ = MPW(x);
      c = x >> MPW_BITS;
    }

    /* --- Get ready for the next round --- */

    av++;
    dv += 2;
  }
}

/* --- @mpx_udiv@ --- *
 *
 * Arguments:	@mpw *qv, *qvl@ = quotient vector base and limit
 *		@mpw *rv, *rvl@ = dividend/remainder vector base and limit
 *		@const mpw *dv, *dvl@ = divisor vector base and limit
 *		@mpw *sv, *svl@ = scratch workspace
 *
 * Returns:	---
 *
 * Use:		Performs unsigned integer division.  If the result overflows
 *		the quotient vector, high-order bits are discarded.  (Clearly
 *		the remainder vector can't overflow.)  The various vectors
 *		may not overlap in any way.  Yes, I know it's a bit odd
 *		requiring the dividend to be in the result position but it
 *		does make some sense really.  The remainder must have
 *		headroom for at least two extra words.  The scratch space
 *		must be at least two words larger than twice the size of the
 *		divisor.
 */

void mpx_udiv(mpw *qv, mpw *qvl, mpw *rv, mpw *rvl,
	      const mpw *dv, const mpw *dvl,
	      mpw *sv, mpw *svl)
{
  unsigned norm = 0;
  size_t scale;
  mpw d, dd;

  /* --- Initialize the quotient --- */

  MPX_ZERO(qv, qvl);

  /* --- Perform some sanity checks --- */

  MPX_SHRINK(dv, dvl);
  assert(((void)"division by zero in mpx_udiv", dv < dvl));

  /* --- Normalize the divisor --- *
   *
   * The algorithm requires that the divisor be at least two digits long.
   * This is easy to fix.
   */

  {
    unsigned b;

    d = dvl[-1];
    for (b = MPW_BITS / 2; b; b >>= 1) {
      if (d < (MPW_MAX >> b)) {
	d <<= b;
	norm += b;
      }
    }
    if (dv + 1 == dvl)
      norm += MPW_BITS;
  }

  /* --- Normalize the dividend/remainder to match --- */

  if (norm) {
    mpw *svvl = sv + (dvl - dv) + 1;
    mpx_lsl(rv, rvl, rv, rvl, norm);
    mpx_lsl(sv, svvl, dv, dvl, norm);
    dv = sv;
    sv = svvl;
    dvl = svvl;
    MPX_SHRINK(dv, dvl);
  }

  MPX_SHRINK(rv, rvl);
  d = dvl[-1];
  dd = dvl[-2];

  /* --- Work out the relative scales --- */

  {
    size_t rvn = rvl - rv;
    size_t dvn = dvl - dv;

    /* --- If the divisor is clearly larger, notice this --- */

    if (dvn > rvn) {
      mpx_lsr(rv, rvl, rv, rvl, norm);
      return;
    }

    scale = rvn - dvn;
  }

  /* --- Calculate the most significant quotient digit --- *
   *
   * Because the divisor has its top bit set, this can only happen once.  The
   * pointer arithmetic is a little contorted, to make sure that the
   * behaviour is defined.
   */

  if (MPX_UCMP(rv + scale, rvl, >=, dv, dvl)) {
    mpx_usub(rv + scale, rvl, rv + scale, rvl, dv, dvl);
    if (qvl - qv > scale)
      qv[scale] = 1;
  }

  /* --- Now for the main loop --- */

  {
    mpw *rvv = rvl - 2;

    while (scale) {
      mpw q;
      mpd rh;

      /* --- Get an estimate for the next quotient digit --- */

      mpw r = rvv[1];
      mpw rr = rvv[0];
      mpw rrr = *--rvv;

      scale--;
      rh = ((mpd)r << MPW_BITS) | rr;
      if (r == d)
	q = MPW_MAX;
      else
	q = MPW(rh / d);

      /* --- Refine the estimate --- */

      {
	mpd yh = (mpd)d * q;
	mpd yl = (mpd)dd * q;

	if (yl > MPW_MAX) {
	  yh += yl >> MPW_BITS;
	  yl &= MPW_MAX;
	}

	while (yh > rh || (yh == rh && yl > rrr)) {
	  q--;
	  yh -= d;
	  if (yl < dd) {
	    yh++;
	    yl += MPW_MAX;
	  }
	  yl -= dd;
	}
      }

      /* --- Remove a chunk from the dividend --- */

      {
	mpw *svv;
	const mpw *dvv;
	mpw c = 0;

	/* --- Calculate the size of the chunk --- */

	for (svv = sv, dvv = dv; dvv < dvl; svv++, dvv++) {
	  mpd x = (mpd)*dvv * (mpd)q + c;
	  *svv = MPW(x);
	  c = x >> MPW_BITS;
	}
	if (c)
	  *svv++ = c;

	/* --- Now make sure that we can cope with the difference --- *
	 *
	 * Take advantage of the fact that subtraction works two's-
	 * complement.
	 */

	mpx_usub(rv + scale, rvl, rv + scale, rvl, sv, svv);
	if (rvl[-1] > MPW_MAX / 2) {
	  mpx_uadd(rv + scale, rvl, rv + scale, rvl, dv, dvl);
	  q--;
	}
      }

      /* --- Done for another iteration --- */

      if (qvl - qv > scale)
	qv[scale] = q;
      r = rr;
      rr = rrr;
    }
  }

  /* --- Now fiddle with unnormalizing and things --- */

  mpx_lsr(rv, rvl, rv, rvl, norm);
}

/*----- That's all, folks -------------------------------------------------*/
